
__author__ = 'Lukas Leufen'
__date__ = '2020-09-21'

import copy
import inspect
import gc
import logging
import os
import pickle
import shutil
from functools import reduce
from typing import Tuple, Union, List
import multiprocessing
import psutil
import dask

import numpy as np
import xarray as xr

from mlair.data_handler.abstract_data_handler import AbstractDataHandler
from mlair.helpers import remove_items, to_list
from mlair.helpers.join import EmptyQueryResult


number = Union[float, int]
num_or_list = Union[number, List[number]]


class DefaultDataHandler(AbstractDataHandler):
    from mlair.data_handler.data_handler_single_station import DataHandlerSingleStation as data_handler
    from mlair.data_handler.data_handler_single_station import DataHandlerSingleStation as data_handler_transformation

    _requirements = remove_items(inspect.getfullargspec(data_handler).args, ["self", "station"])

    DEFAULT_ITER_DIM = "Stations"
    DEFAULT_TIME_DIM = "datetime"

    def __init__(self, id_class: data_handler, experiment_path: str, min_length: int = 0,
                 extreme_values: num_or_list = None, extremes_on_right_tail_only: bool = False, name_affix=None,
                 store_processed_data=True, iter_dim=DEFAULT_ITER_DIM, time_dim=DEFAULT_TIME_DIM):
        super().__init__()
        self.id_class = id_class
        self.time_dim = time_dim
        self.iter_dim = iter_dim
        self.min_length = min_length
        self._X = None
        self._Y = None
        self._X_extreme = None
        self._Y_extreme = None
        _name_affix = str(f"{str(self.id_class)}_{name_affix}" if name_affix is not None else id(self))
        self._save_file = os.path.join(experiment_path, "data", f"{_name_affix}.pickle")
        self._collection = self._create_collection()
        self.harmonise_X()
        self.multiply_extremes(extreme_values, extremes_on_right_tail_only, dim=self.time_dim)
        self._store(fresh_store=True, store_processed_data=store_processed_data)

    @classmethod
    def build(cls, station: str, **kwargs):
        sp_keys = {k: copy.deepcopy(kwargs[k]) for k in cls._requirements if k in kwargs}
        sp = cls.data_handler(station, **sp_keys)
        dp_args = {k: copy.deepcopy(kwargs[k]) for k in cls.own_args("id_class") if k in kwargs}
        return cls(sp, **dp_args)

    def _create_collection(self):
        return [self.id_class]

    @classmethod
    def requirements(cls):
        return remove_items(super().requirements(), "id_class")

    def _reset_data(self):
        self._X, self._Y, self._X_extreme, self._Y_extreme = None, None, None, None
        gc.collect()

    def _cleanup(self):
        directory = os.path.dirname(self._save_file)
        if os.path.exists(directory) is False:
            os.makedirs(directory)
        if os.path.exists(self._save_file):
            shutil.rmtree(self._save_file, ignore_errors=True)

    def _store(self, fresh_store=False, store_processed_data=True):
        if store_processed_data is True:
            self._cleanup() if fresh_store is True else None
            data = {"X": self._X, "Y": self._Y, "X_extreme": self._X_extreme, "Y_extreme": self._Y_extreme}
            with open(self._save_file, "wb") as f:
                pickle.dump(dask.compute(data), f)
            logging.debug(f"save pickle data to {self._save_file}")
            self._reset_data()

    def _load(self):
        try:
            with open(self._save_file, "rb") as f:
                data = pickle.load(f)
            logging.debug(f"load pickle data from {self._save_file}")
            self._X, self._Y = data["X"], data["Y"]
            self._X_extreme, self._Y_extreme = data["X_extreme"], data["Y_extreme"]
        except FileNotFoundError:
            pass

    def get_data(self, upsampling=False, as_numpy=True):
        self._load()
        X = self.get_X(upsampling, as_numpy)
        Y = self.get_Y(upsampling, as_numpy)
        self._reset_data()
        return X, Y

    def __repr__(self):
        return ";".join(list(map(lambda x: str(x), self._collection)))

    def get_X_original(self):
        X = []
        for data in self._collection:
            X.append(data.get_X())
        return X

    def get_Y_original(self):
        Y = self._collection[0].get_Y()
        return Y

    @staticmethod
    def _to_numpy(d):
        return list(map(lambda x: np.copy(x), d))

    def get_X(self, upsampling=False, as_numpy=True):
        no_data = (self._X is None)
        self._load() if no_data is True else None
        X = self._X if upsampling is False else self._X_extreme
        self._reset_data() if no_data is True else None
        return self._to_numpy(X) if as_numpy is True else X

    def get_Y(self, upsampling=False, as_numpy=True):
        no_data = (self._Y is None)
        self._load() if no_data is True else None
        Y = self._Y if upsampling is False else self._Y_extreme
        self._reset_data() if no_data is True else None
        return self._to_numpy([Y]) if as_numpy is True else Y

    def harmonise_X(self):
        X_original, Y_original = self.get_X_original(), self.get_Y_original()
        dim = self.time_dim
        intersect = reduce(np.intersect1d, map(lambda x: x.coords[dim].values, X_original))
        if len(intersect) < max(self.min_length, 1):
            X, Y = None, None
        else:
            X = list(map(lambda x: x.sel({dim: intersect}), X_original))
            Y = Y_original.sel({dim: intersect})
        self._X, self._Y = X, Y

    def get_observation(self):
        return self.id_class.observation.copy().squeeze()

    def apply_transformation(self, data, base="target", dim=0, inverse=False):
        return self.id_class.apply_transformation(data, dim=dim, base=base, inverse=inverse)

    def multiply_extremes(self, extreme_values: num_or_list = 1., extremes_on_right_tail_only: bool = False,
                          timedelta: Tuple[int, str] = (1, 'm'), dim=DEFAULT_TIME_DIM):
        """
        Multiply extremes.

        This method extracts extreme values from self.labels which are defined in the argument extreme_values. One can
        also decide only to extract extremes on the right tail of the distribution. When extreme_values is a list of
        floats/ints all values larger (and smaller than negative extreme_values; extraction is performed in standardised
        space) than are extracted iteratively. If for example extreme_values = [1.,2.] then a value of 1.5 would be
        extracted once (for 0th entry in list), while a 2.5 would be extracted twice (once for each entry). Timedelta is
        used to mark those extracted values by adding one min to each timestamp. As TOAR Data are hourly one can
        identify those "artificial" data points later easily. Extreme inputs and labels are stored in
        self.extremes_history and self.extreme_labels, respectively.

        :param extreme_values: user definition of extreme
        :param extremes_on_right_tail_only: if False also multiply values which are smaller then -extreme_values,
            if True only extract values larger than extreme_values
        :param timedelta: used as arguments for np.timedelta in order to mark extreme values on datetime
        """
        # check if X or Y is None
        if (self._X is None) or (self._Y is None):
            logging.debug(f"{str(self.id_class)} has no data for X or Y, skip multiply extremes")
            return
        if extreme_values is None:
            logging.debug(f"No extreme values given, skip multiply extremes")
            self._X_extreme, self._Y_extreme = self._X, self._Y
            return

        # check type if inputs
        extreme_values = to_list(extreme_values)
        for i in extreme_values:
            if not isinstance(i, number.__args__):
                raise TypeError(f"Elements of list extreme_values have to be {number.__args__}, but at least element "
                                f"{i} is type {type(i)}")

        for extr_val in sorted(extreme_values):
            # check if some extreme values are already extracted
            if (self._X_extreme is None) or (self._Y_extreme is None):
                X = self._X
                Y = self._Y
            else:  # one extr value iteration is done already: self.extremes_label is NOT None...
                X = self._X_extreme
                Y = self._Y_extreme

            # extract extremes based on occurrence in labels
            other_dims = remove_items(list(Y.dims), dim)
            if extremes_on_right_tail_only:
                extreme_idx = (Y > extr_val).any(dim=other_dims)
            else:
                extreme_idx = xr.concat([(Y < -extr_val).any(dim=other_dims[0]),
                                           (Y > extr_val).any(dim=other_dims[0])],
                                          dim=other_dims[1]).any(dim=other_dims[1])

            extremes_X = list(map(lambda x: x.sel(**{dim: extreme_idx}), X))
            self._add_timedelta(extremes_X, dim, timedelta)
            # extremes_X = list(map(lambda x: x.coords[dim].values + np.timedelta64(*timedelta), extremes_X))

            extremes_Y = Y.sel(**{dim: extreme_idx})
            extremes_Y.coords[dim].values += np.timedelta64(*timedelta)

            self._Y_extreme = xr.concat([Y, extremes_Y], dim=dim)
            self._X_extreme = list(map(lambda x1, x2: xr.concat([x1, x2], dim=dim), X, extremes_X))

    @staticmethod
    def _add_timedelta(data, dim, timedelta):
        for d in data:
            d.coords[dim].values += np.timedelta64(*timedelta)

    @classmethod
    def transformation(cls, set_stations, **kwargs):
        """
        ### supported transformation methods

        Currently supported methods are:

        * standardise (default, if method is not given)
        * centre

        ### mean and std estimation

        Mean and std (depending on method) are estimated. For each station, mean and std are calculated and afterwards
        aggregated using the mean value over all station-wise metrics. This method is not exactly accurate, especially
        regarding the std calculation but therefore much faster. Furthermore, it is a weighted mean weighted by the
        time series length / number of data itself - a longer time series has more influence on the transformation
        settings than a short time series. The estimation of the std in less accurate, because the unweighted mean of
        all stds in not equal to the true std, but still the mean of all station-wise std is a decent estimate. Finally,
        the real accuracy of mean and std is less important, because it is "just" a transformation / scaling.

        ### mean and std given

        If mean and std are not None, the default data handler expects this parameters to match the data and applies
        this values to the data. Make sure that all dimensions and/or coordinates are in agreement.
        """

        sp_keys = {k: copy.deepcopy(kwargs[k]) for k in cls._requirements if k in kwargs}
        transformation_dict = sp_keys.get("transformation", None)
        if transformation_dict is None:
            return
        if isinstance(transformation_dict, dict):  # tuple for (input, target) transformation
            transformation_dict = copy.deepcopy(transformation_dict), copy.deepcopy(transformation_dict)

        def _inner():
            """Inner method that is performed in both serial and parallel approach."""
            if dh is not None:
                for i, transformation in enumerate(dh._transformation):
                    for var in transformation.keys():
                        if var not in transformation_dict[i].keys():
                            transformation_dict[i][var] = {}
                        opts = transformation[var]
                        assert transformation_dict[i][var].get("method", opts["method"]) == opts["method"]
                        transformation_dict[i][var]["method"] = opts["method"]
                        for k in ["mean", "std", "min", "max"]:
                            old = transformation_dict[i][var].get(k, None)
                            new = opts.get(k)
                            transformation_dict[i][var][k] = new if old is None else old.combine_first(new)

        if multiprocessing.cpu_count() > 1:  # parallel solution
            logging.info("use parallel transformation approach")
            pool = multiprocessing.Pool(
                min([psutil.cpu_count(logical=False), len(set_stations), 16]))  # use only physical cpus
            logging.info(f"running {getattr(pool, '_processes')} processes in parallel")
            output = [
                pool.apply_async(f_proc, args=(cls.data_handler_transformation, station), kwds=sp_keys)
                for station in set_stations]
            for p in output:
                dh, s = p.get()
                _inner()
        else:  # serial solution
            logging.info("use serial transformation approach")
            for station in set_stations:
                dh, s = f_proc(cls.data_handler_transformation, station, **sp_keys)
                _inner()

        # aggregate all information
        iter_dim = sp_keys.get("iter_dim", cls.DEFAULT_ITER_DIM)
        pop_list = []
        for i, transformation in enumerate(transformation_dict):
            for k in transformation.keys():
                try:
                    if transformation[k]["mean"] is not None:
                        transformation_dict[i][k]["mean"] = transformation[k]["mean"].mean(iter_dim)
                    if transformation[k]["std"] is not None:
                        transformation_dict[i][k]["std"] = transformation[k]["std"].mean(iter_dim)
                    if transformation[k]["min"] is not None:
                        transformation_dict[i][k]["min"] = transformation[k]["min"].min(iter_dim)
                    if transformation[k]["max"] is not None:
                        transformation_dict[i][k]["max"] = transformation[k]["max"].max(iter_dim)
                except KeyError:
                    pop_list.append((i, k))
        for (i, k) in pop_list:
            transformation_dict[i].pop(k)
        return transformation_dict

    def get_coordinates(self):
        return self.id_class.get_coordinates()


def f_proc(data_handler, station, **sp_keys):
    """
    Try to create a data handler for given arguments. If build fails, this station does not fulfil all requirements and
    therefore f_proc will return None as indication. On a successful build, f_proc returns the built data handler and
    the station that was used. This function must be implemented globally to work together with multiprocessing.
    """
    try:
        res = data_handler(station, **sp_keys)
    except (AttributeError, EmptyQueryResult, KeyError, ValueError) as e:
        logging.info(f"remove station {station} because it raised an error: {e}")
        res = None
    return res, station
