"""Collection of different help functions."""
__author__ = 'Lukas Leufen, Felix Kleinert'
__date__ = '2019-10-21'

import inspect
import math

import numpy as np
import xarray as xr
import dask.array as da

from typing import Dict, Callable, Union, List, Any


def to_list(obj: Any) -> List:
    """
    Transform given object to list if obj is not already a list. Sets are also transformed to a list.

    :param obj: object to transform to list

    :return: list containing obj, or obj itself (if obj was already a list)
    """
    if isinstance(obj, (set, tuple)):
        obj = list(obj)
    elif not isinstance(obj, list):
        obj = [obj]
    return obj


def dict_to_xarray(d: Dict, coordinate_name: str) -> xr.DataArray:
    """
    Convert a dictionary of 2D-xarrays to single 3D-xarray. The name of new coordinate axis follows <coordinate_name>.

    :param d: dictionary with 2D-xarrays
    :param coordinate_name: name of the new created axis (2D -> 3D)

    :return: combined xarray
    """
    if len(d.keys()) == 1:
        k = list(d.keys())
        xarray: xr.DataArray = d[k[0]]
        return xarray.expand_dims(dim={coordinate_name: k}, axis=0)
    else:
        xarray = None
        for k, v in d.items():
            if xarray is None:
                xarray = v
                xarray.coords[coordinate_name] = k
            else:
                tmp_xarray = v
                tmp_xarray.coords[coordinate_name] = k
                xarray = xr.concat([xarray, tmp_xarray], coordinate_name)
        return xarray


def float_round(number: float, decimals: int = 0, round_type: Callable = math.ceil) -> float:
    """
    Perform given rounding operation on number with the precision of decimals.

    :param number: the number to round
    :param decimals: numbers of decimals of the rounding operations (default 0 -> round to next integer value)
    :param round_type: the actual rounding operation. Can be any callable function like math.ceil, math.floor or python
        built-in round operation.

    :return: rounded number with desired precision
    """
    multiplier = 10. ** decimals
    return round_type(number * multiplier) / multiplier


def remove_items(obj: Union[List, Dict], items: Any):
    """
    Remove item(s) from either list or dictionary.

    :param obj: object to remove items from (either dictionary or list)
    :param items: elements to remove from obj. Can either be a list or single entry / key

    :return: object without items
    """

    def remove_from_list(list_obj, item_list):
        """Remove implementation for lists."""
        if len(items) > 1:
            return [e for e in list_obj if e not in item_list]
        else:
            list_obj = list_obj.copy()
            try:
                list_obj.remove(item_list[0])
            except ValueError:
                pass
            return list_obj

    def remove_from_dict(dict_obj, key_list):
        """Remove implementation for dictionaries."""
        return {k: v for k, v in dict_obj.items() if k not in key_list}

    items = to_list(items)
    if isinstance(obj, list):
        return remove_from_list(obj, items)
    elif isinstance(obj, dict):
        return remove_from_dict(obj, items)
    else:
        raise TypeError(f"{inspect.stack()[0][3]} does not support type {type(obj)}.")


def select_from_dict(dict_obj: dict, sel_list: Any):
    """
    Extract all key values pairs whose key is contained in the sel_list.

    Does not perform a check if all elements of sel_list are keys of dict_obj. Therefore the number of pairs in the
    returned dict is always smaller or equal to the number of elements in the sel_list.
    """
    sel_list = to_list(sel_list)
    assert isinstance(dict_obj, dict)
    sel_dict = {k: v for k, v in dict_obj.items() if k in sel_list}
    return sel_dict


def extract_value(encapsulated_value):
    try:
        if isinstance(encapsulated_value, str):
            raise TypeError
        if len(encapsulated_value) == 1:
            return extract_value(encapsulated_value[0])
        else:
            raise NotImplementedError("Trying to extract an encapsulated value from objects with more than a single "
                                      "entry is not supported by this function.")
    except TypeError:
        return encapsulated_value


def is_xarray(arr) -> bool:
    """
    Returns True if arr is xarray.DataArray or xarray.Dataset.
    :param arr: variable in question
    :type arr: Any
    :return:
    :rtype: bool
    """
    return isinstance(arr, xr.DataArray) or isinstance(arr, xr.Dataset)


def convert2xrda(arr: Union[xr.DataArray, xr.Dataset, np.ndarray, int, float],
                 use_1d_default: bool = False, **kwargs) -> Union[xr.DataArray, xr.Dataset]:
    """
    Converts np.array, int or float object to xarray.DataArray.

    If a xarray.DataArray or xarray.Dataset is passed, returns that unchanged.
    :param arr:
    :type arr: xr.DataArray, xr.Dataset, np.ndarray, int, float
    :param use_1d_default:
    :type use_1d_default: bool
    :param kwargs: Any additional kwargs which are accepted by xr.DataArray()
    :type kwargs:
    :return:
    :rtype: xr.DataArray, xr.DataSet
    """
    if is_xarray(arr):
        return arr
    else:
        if use_1d_default:
            if isinstance(arr, da.core.Array):
                raise TypeError(f"`use_1d_default=True' is used with `arr' of type da.array. For da.arrays please "
                                f"pass `use_1d_default=False' and specify keywords for xr.DataArray via kwargs.")
            dims = kwargs.pop('dims', 'points')
            coords = kwargs.pop('coords', None)
            try:
                if coords is None:
                    coords = {dims: range(arr.shape[0])}
            except (AttributeError, IndexError):
                if isinstance(arr, int) or isinstance(arr, float):
                    coords = kwargs.pop('coords', {dims: range(1)})
                    dims = to_list(dims)
                else:
                    raise TypeError(f"`arr' must be arry-like, int or float. But is of type {type(arr)}")
            kwargs.update({'dims': dims, 'coords': coords})

        return xr.DataArray(arr, **kwargs)


